/*
 * extract_xarc.c
 *
 * マクロREPLACE_XARCを定義すると置換モード
 * 未定義のときは展開モード
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

#include "utils.h"

#ifdef WIN32
#include "my_printf.h"
#define printf		printf_utf8
#endif


#ifdef REPLACE_XARC
static uint32_t WARN_FILE_SIZE = 0x7FFFFFFF;
static uint32_t MAX_FILE_SIZE  = 0xFFFFFFFF;
#else	/* REPLACE_XARC */
static int MODE_LIST    = 0;
static int MODE_EXTRACT = 1;
#endif	/* REPLACE_XARC */

struct xarc_info_t {
	unsigned char	*data;			/* アーカイブデータ */
	size_t			size;			/* アーカイブサイズ */

	uint32_t		unknown;
	uint32_t		num_files;
	uint32_t		data_offset;

	struct xarc_file_entry_t {
		size_t			entry_offset;
		size_t			data_offset;

		char			*filename;
		uint32_t		size;
		uint32_t		unknown;

#ifdef REPLACE_XARC
		char			*new_file;
		size_t			new_size;
		unsigned char	*new_data;
#endif	/* REPLACE_XARC */
	} *file_entries;
};


static void usage(void);
#ifdef REPLACE_XARC
static int replace_xarc(const char *filename, const char *out_file, int num_replace_files, char **replace_files);
#else	/* REPLACE_XARC */
static int extract_xarc(const char *filename, const char *out_dir, int mode, const char *ext);
#endif	/* REPLACE_XARC */
#if 0
static int get_uint32(struct xarc_info_t *info, size_t *offset, uint32_t *value);
#endif
#ifdef REPLACE_XARC
static int read_replace_files(struct xarc_info_t *info, int num_replace_files, char **replace_files);
static struct xarc_file_entry_t *search_xarc_file_entry(struct xarc_info_t *info, const char *filename);
static int write_xarc_file(const char *filename, const struct xarc_info_t *info);
#endif	/* REPLACE_XARC */


#ifdef REPLACE_XARC
static int opt_warn_large_file = 0;
#endif	/* REPLACE_XARC */


int main(int argc, char **argv)
{
	const char *xarc_file = NULL;
#ifdef REPLACE_XARC
	const char *out_file = NULL;
	int num_replace_files = 0;
	char **replace_files = NULL;
#else	/* REPLACE_XARC */
	const char *out_dir = NULL;
	const char *ext = NULL;
	int mode = MODE_EXTRACT;
#endif	/* REPLACE_XARC */

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
#ifdef REPLACE_XARC
			} else if (strcmp(*argv, "-o") == 0 || strcmp(*argv, "--output") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 出力ファイル名がありません\n");
					return 1;
				}
				out_file = *argv;
			} else if (strcmp(*argv, "-W") == 0 || strcmp(*argv, "--warn-large-file") == 0) {
				opt_warn_large_file = 1;
#else	/* REPLACE_XARC */
			} else if (strcmp(*argv, "-l") == 0 || strcmp(*argv, "--list") == 0) {
				mode = MODE_LIST;
			} else if (strcmp(*argv, "-x") == 0 || strcmp(*argv, "--extract") == 0) {
				mode = MODE_EXTRACT;
			} else if (strcmp(*argv, "-e") == 0 || strcmp(*argv, "--ext") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 拡張子がありません\n");
					return 1;
				}
				ext = *argv;
#endif	/* REPLACE_XARC */
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (xarc_file == NULL) {
				xarc_file = *argv;
#ifdef REPLACE_XARC
			} else {
				size_t alloc_size = sizeof(char *) * (num_replace_files + 1);
				if (num_replace_files == 0) {
					replace_files = (char **)malloc(alloc_size);
				} else {
					replace_files = (char **)realloc(replace_files, alloc_size);
				}
				if (replace_files == NULL) {
					printf("ERROR: memory exhausted!\n");
					return 1;
				}
				replace_files[num_replace_files] = *argv;
				convert_backslash_to_slash(replace_files[num_replace_files]);
				num_replace_files++;
#else	/* REPLACE_XARC */
			} else if (out_dir == NULL) {
				out_dir = *argv;
			} else {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
#endif	/* REPLACE_XARC */
			}
		}
	}

	if (xarc_file == NULL) {
#if 1
		usage();
#else
		printf("ERROR: アーカイブファイルが指定されていません\n");
#endif
		return 1;
	}

#ifndef REPLACE_XARC
	if (mode == MODE_LIST && out_dir != NULL) {
		printf("WARNING: 一覧表示モードのときには展開先ディレクトリは使用しません: %s\n", out_dir);
	}
#endif	/* not REPLACE_XARC */


	printf("アーカイブファイル: %s\n", xarc_file);
#ifdef REPLACE_XARC
	if (num_replace_files == 0) {
		printf("ERROR: 置換ファイルが一つも指定されていません\n");
		return 1;
	}
	if (out_file == NULL &&
		(out_file = create_new_filename(xarc_file)) == NULL) {
		printf("ERROR: 出力ファイル名を作成できません\n");
		return 1;
	}
	printf("出力ファイル      : %s\n", out_file);
	if (replace_xarc(xarc_file, out_file, num_replace_files, replace_files) < 0) {
		printf("ERROR: 置換できませんでした: %s\n", xarc_file);
	}
#else	/* REPLACE_XARC */
	if (mode == MODE_EXTRACT) {
		printf("展開先ディレクトリ: %s\n", out_dir == NULL ? "." : out_dir);
	}
	if (extract_xarc(xarc_file, out_dir, mode, ext) < 0) {
		printf("ERROR: %sできませんでした: %s\n", (mode == MODE_LIST ? "一覧表示" : "展開"), xarc_file);
		return 1;
	}
#endif	/* REPLACE_XARC */
	printf("処理完了\n\n");

	return 0;
}

static void usage(void)
{
#ifdef REPLACE_XARC
	printf("使用方法: replace_xarc [オプション] xarcファイル 置換ファイル...\n");
	printf("機    能: xarcアーカイブ内のファイルを入れ替えます\n");
	printf("オプション:\n");
	printf("    -o|--output ファイル\n");
    printf("                        指定したファイルに新しいxarcアーカイブを出力します\n");
	printf("                        指定の無いときは元のファイル名の拡張子の前に .NEW を\n");
	printf("                        挿入した名前で作成します\n");
	printf("    -W|--warn-large-file\n");
	printf("                        置換ファイルが%u(0x%X)バイトを越える\n", WARN_FILE_SIZE, WARN_FILE_SIZE);
	printf("                        ときに警告表示します\n");
	printf("                        置換可能な最大サイズは%u(0x%X)バイトです\n", MAX_FILE_SIZE, MAX_FILE_SIZE);
	printf("    -?|--help           使用方法を表示します\n");
	printf("\n");
	printf("xarcアーカイブ内のファイルを同じ名前の置換ファイルで置き替えます\n");
	printf("同じ名前のファイルが複数ある場合は先に指定されたものを使用します\n");
	printf("ディレクトリを指定した場合はその中のファイルを置換ファイルとします\n");
	printf("アーカイブ内のファイル名と置換ファイル名の比較時には大文字小文字を区別しません\n");
	printf("アーカイブ内に無いファイルは無視します\n");
#else	/* REPLACE_XARC */
	printf("使用方法: extract_xarc [オプション] xarcファイル [展開先ディレクトリ]\n");
	printf("機    能: xarcアーカイブを展開します\n");
	printf("オプション:\n");
	printf("    -l|--list           アーカイブ内容を一覧表示します\n");
	printf("                        ファイル展開はしません\n");
	printf("    -x|--extract        アーカイブ内のファイルを展開します(デフォルト)\n");
	printf("    -e|--ext 拡張子     指定した拡張子のファイルのみ展開します\n");
	printf("    -?|--help           使用方法を表示します\n");
#endif	/* REPLACE_XARC */
}

#ifdef REPLACE_XARC
static int replace_xarc(const char *filename, const char *out_file, int num_replace_files, char **replace_files)
#else	/* REPLACE_XARC */
static int extract_xarc(const char *filename, const char *out_dir, int mode, const char *ext)
#endif	/* REPLACE_XARC */
{
	int result = -1;
	struct xarc_info_t info;
	size_t offset = 0;
	size_t data_offset = 0;
	size_t alloc_size = 0;
	int i;
#ifdef REPLACE_XARC
	int num_new_files = 0;
#endif	/* REPLACE_XARC */

	memset(&info, 0, sizeof(info));

	if (read_file(filename, &(info.data), &(info.size)) < 0) {
		printf("ERROR: ファイルを読み込めません: %s\n", filename);
		goto ERROR_EXIT;
	}

#ifndef REPLACE_XARC
	if (mode == MODE_EXTRACT && out_dir != NULL && *out_dir != '\0') {
		struct stat st;

		if (stat(out_dir, &st) < 0 && errno == ENOENT) {
			if (make_directory(out_dir, 0) < 0) {
				printf("WARNING: 出力ディレクトリを作成できません: %s\n", out_dir);
			}
		}
		if (chdir(out_dir) < 0) {
			printf("ERROR: ディレクトリを変更できません: %s\n", out_dir);
			goto ERROR_EXIT;
		}
	}
#endif	/* not REPLACE_XARC */

	if (get_uint32(info.data, &offset, info.size, &(info.unknown)) < 0) {
		printf("ERROR: unknown valueを取得できません\n");
		goto ERROR_EXIT;
	}

	if (get_uint32(info.data, &offset, info.size, &(info.num_files)) < 0) {
		printf("ERROR: ファイル数を取得できません\n");
		goto ERROR_EXIT;
	}

	if (get_uint32(info.data, &offset, info.size, &(info.data_offset)) < 0) {
		printf("ERROR: データオフセットを取得できません\n");
		goto ERROR_EXIT;
	}

	alloc_size = sizeof(struct xarc_file_entry_t) * info.num_files;
	if ((info.file_entries = (struct xarc_file_entry_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	memset(info.file_entries, 0, alloc_size);

	data_offset = info.data_offset;
	for (i = 0; i < info.num_files; i++) {
		info.file_entries[i].entry_offset = offset;
		info.file_entries[i].filename = (char *)info.data + offset;
		while (1) {
			if (info.size <= offset) {
				printf("ERROR: ファイル名終端を見つけられません: index=%d, offset=%08zX\n", i, info.file_entries[i].entry_offset);
				goto ERROR_EXIT;
			}
			if (info.data[offset++] == 0) {
				break;
			}
		}

		if (get_uint32(info.data, &offset, info.size, &(info.file_entries[i].size)) < 0) {
			printf("ERROR: ファイルサイズを取得できません: index=%d, offset=%08zX\n", i, info.file_entries[i].entry_offset);
			goto ERROR_EXIT;
		}
		if (info.size < data_offset + info.file_entries[i].size) {
			printf("ERROR: ファイルサイズが大きすぎます: index=%d, offset=%08zX\n", i, info.file_entries[i].entry_offset);
			goto ERROR_EXIT;
		}

		if (get_uint32(info.data, &offset, info.size, &(info.file_entries[i].unknown)) < 0) {
			printf("ERROR: unknown valueを取得できません: index=%d, offset=%08zX\n", i, info.file_entries[i].entry_offset);
			goto ERROR_EXIT;
		}

		if (info.data_offset < offset) {
			printf("ERROR: ファイルエントリが大きすぎます: index=%d, offset=%08zX, data_offset=%08X\n", i, info.file_entries[i].entry_offset, info.data_offset);
			goto ERROR_EXIT;
		}

		info.file_entries[i].data_offset = data_offset;
		data_offset += info.file_entries[i].size;

#ifndef REPLACE_XARC
		if (i == 0) {
			printf("no   offset   size(h)  size       filename\n");
			printf("---- -------- -------- ---------- ----------------------------------------\n");
		}

		if (ext != NULL) {
			/* 指定された拡張子以外はスキップ */
			/* ext == "" のときは拡張子なしのファイルを対象とする */
			const char *ptr = strrchr(info.file_entries[i].filename, '.');
			if (ptr == NULL) {
				if (*ext != '\0') {
					continue;
				}
			} else {
				ptr++;
				if (strcasecmp(ptr, ext) != 0) {
					continue;
				}
			}
		}

		printf("%04d %08zX %08X %10d %s\n", i, info.file_entries[i].data_offset, info.file_entries[i].size, info.file_entries[i].size, info.file_entries[i].filename);
		if (mode == MODE_EXTRACT) {
			if (write_file(info.file_entries[i].filename, info.data + info.file_entries[i].data_offset, info.file_entries[i].size) < 0) {
				printf("ERROR: ファイルを書き込めません: %s\n", info.file_entries[i].filename);
				goto ERROR_EXIT;
			}
		}
#endif	/* not REPLACE_XARC */
	}

	if (data_offset < info.size) {
		printf("WARNING: 未使用データがあります: offset=%08zX\n", data_offset);
	}

#ifdef REPLACE_XARC
	if ((num_new_files = read_replace_files(&info, num_replace_files, replace_files)) < 0) {
		printf("ERROR: 置換ファイルを読み込めません\n");
		goto ERROR_EXIT;
	}

	if (num_new_files == 0) {
		printf("ERROR: 置換できるファイルがありません\n");
		goto ERROR_EXIT;
	}

	if (write_xarc_file(out_file, &info) < 0) {
		printf("ERROR: ファイルを書き込めません: %s\n", out_file);
		goto ERROR_EXIT;
	}
#endif	/* REPLACE_XARC */

	result = 0;

ERROR_EXIT:

	return result;
}

#if 0
static int get_uint32(struct xarc_info_t *info, size_t *offset, uint32_t *value)
{
	const unsigned char *data;

	if (info->size < *offset + 4) {
		return -1;
	}

	data = info->data + *offset;
	*value = ((uint32_t)*data) | (((uint32_t)*(data + 1)) << 8) | (((uint32_t)*(data + 2)) << 16) | (((uint32_t)*(data + 3)) << 24);
	*offset += 4;

	return 0;
}
#endif

#ifdef REPLACE_XARC
static int read_replace_files(struct xarc_info_t *info, int num_replace_files, char **replace_files)
{
	int result = -1;
	DIR *dirp = NULL;
	int i;
	struct stat st;
	int num_new_files = 0;

	for (i = 0; i < num_replace_files; i++) {
		if (stat(replace_files[i], &st) < 0) {
			printf("ERROR: %d: ファイル情報を取得できません: %s\n", errno, replace_files[i]);
			goto ERROR_EXIT;
		}

		if (S_ISDIR(st.st_mode)) {
			struct dirent *dp = NULL;
			struct xarc_file_entry_t *entry = NULL;

			if ((dirp = opendir(replace_files[i])) == NULL) {
				printf("ERROR: ディレクトリを開けません: %s\n", replace_files[i]);
				goto ERROR_EXIT;
			}

			while ((dp = readdir(dirp)) != NULL) {
				if ((entry = search_xarc_file_entry(info, dp->d_name)) != NULL) {
					break;
				}
			}

			if (entry != NULL) {
				size_t alloc_size = strlen(replace_files[i]) + 1 + strlen(dp->d_name) + 1;
				if ((entry->new_file = (char *)malloc(alloc_size)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}
				strcat(strcat(strcpy(entry->new_file, replace_files[i]), "/"), dp->d_name);

				printf("置換ファイル      : %s\n", entry->new_file);
				if (read_file(entry->new_file, &(entry->new_data), &(entry->new_size)) < 0) {
					printf("ERROR: ファイルを読めません: %s\n", entry->new_file);
					goto ERROR_EXIT;
				}

				if (MAX_FILE_SIZE < entry->new_size ||
					(WARN_FILE_SIZE < entry->new_size && opt_warn_large_file != 0)) {
					if (MAX_FILE_SIZE < entry->new_size) {
						printf("ERROR: ファイルが大きすぎます: 最大%u(0x%X)バイト", MAX_FILE_SIZE, MAX_FILE_SIZE);
					} else {
						printf("WARNING: ファイルサイズが%u(0x%X)バイトを越えています", WARN_FILE_SIZE, WARN_FILE_SIZE);
					}
					printf(": %s: %zu(0x%zX)バイト\n", entry->new_file, entry->new_size, entry->new_size);

					if (MAX_FILE_SIZE < entry->new_size) {
						goto ERROR_EXIT;
					}
				}

				num_new_files++;
			}

			closedir(dirp);
			dirp = NULL;
		} else {
			struct xarc_file_entry_t *entry;

			if ((entry = search_xarc_file_entry(info, replace_files[i])) == NULL) {
				continue;
			}
			entry->new_file = replace_files[i];

			printf("置換ファイル      : %s\n", entry->new_file);
			if (read_file(entry->new_file, &(entry->new_data), &(entry->new_size)) < 0) {
				printf("ERROR: ファイルを読めません: %s\n", entry->new_file);
				goto ERROR_EXIT;
			}

			if (MAX_FILE_SIZE < entry->new_size ||
				(WARN_FILE_SIZE < entry->new_size && opt_warn_large_file != 0)) {
				if (MAX_FILE_SIZE < entry->new_size) {
					printf("ERROR: ファイルが大きすぎます: 最大%u(0x%X)バイト", MAX_FILE_SIZE, MAX_FILE_SIZE);
				} else {
					printf("WARNING: ファイルサイズが%u(0x%X)バイトを越えています", WARN_FILE_SIZE, WARN_FILE_SIZE);
				}
				printf(": %s: %zu(0x%zX)バイト\n", entry->new_file, entry->new_size, entry->new_size);

				if (MAX_FILE_SIZE < entry->new_size) {
					goto ERROR_EXIT;
				}
			}

			num_new_files++;
		}
	}

	result = num_new_files;

ERROR_EXIT:
	if (dirp != NULL) {
		closedir(dirp);
	}

	return result;
}

static struct xarc_file_entry_t *search_xarc_file_entry(struct xarc_info_t *info, const char *filename)
{
	const char *ptr;
	int i;

	if ((ptr = strrchr(filename, '/')) != NULL) {
		filename = ptr + 1;
	}

	for (i = 0; info->num_files; i++) {
		if (info->file_entries[i].new_file == NULL &&
			strcasecmp(info->file_entries[i].filename, filename) == 0) {
			return &(info->file_entries[i]);
		}
	}

	return NULL;
}

static int write_xarc_file(const char *filename, const struct xarc_info_t *info)
{
	int result = -1;
	FILE *fp = NULL;
	size_t offset = 0;
	int i;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (write_uint32(fp, info->unknown) < 0 ||
		write_uint32(fp, info->num_files) < 0 ||
		write_uint32(fp, info->data_offset) < 0) {
		goto ERROR_EXIT;
	}
	offset = 12;

	/* ファイルエントリ */
	for (i = 0; i < info->num_files; i++) {
		uint32_t size = 0;
		int name_len = strlen(info->file_entries[i].filename);

		if (info->file_entries[i].new_file == NULL) {
			size = info->file_entries[i].size;
		} else {
			size = (uint32_t)info->file_entries[i].new_size;
		}

		if (write_buffer(fp, (const unsigned char *)info->file_entries[i].filename, name_len + 1) < 0 ||
			write_uint32(fp, size) < 0 ||
			write_uint32(fp, info->file_entries[i].unknown) < 0) {
			goto ERROR_EXIT;
		}
		offset += name_len + 1 + 4 + 4;
	}

	if (offset != info->data_offset) {
		printf("BUG: データオフセットが変化しました: orig=%08X, new=%08zX\n", info->data_offset, offset);
		goto ERROR_EXIT;
	}

	/* ファイルデータ */
	for (i = 0; i < info->num_files; i++) {
		uint32_t size = 0;
		const unsigned char *data = NULL;

		if (info->file_entries[i].new_file == NULL) {
			size = info->file_entries[i].size;
			data = info->data + info->file_entries[i].data_offset;
		} else {
			size = (uint32_t)info->file_entries[i].new_size;
			data = info->file_entries[i].new_data;
		}

		if (write_buffer(fp, data, size) < 0) {
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}
#endif	/* REPLACE_XARC */

/* EOF */
