#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <iconv.h>
#include <errno.h>

#include "my_printf.h"


#ifdef WIN32
static void check_lang(void);
#endif
static char *make_text(const char *format, va_list ap);
static int convert_and_print_text(const char *to_encoding, const char *from_encoding, const char *from_encoding2, char *text);


#ifdef WIN32
static int lang_checked = 0;
static int utf8_conversion = 1;
static int cp932_conversion = 0;
static const char *latin1_to_encoding = "CP932";
#else
static const char *latin1_to_encoding = "UTF-8";
#endif


#ifdef WIN32
int printf_utf8(const char *format, ...)
{
	int result = -1;
	va_list ap;
	char *text = NULL;

	if (lang_checked == 0) {
		check_lang();
	}

	va_start(ap, format);

	if (utf8_conversion == 0) {
		vprintf(format, ap);
	} else {
		if ((text = make_text(format, ap)) == NULL) {
			goto ERROR_EXIT;
		}

		if (convert_and_print_text("CP932", "UTF-8", NULL, text) < 0) {
			goto ERROR_EXIT;
		}
	}

	va_end(ap);

	result = 0;

ERROR_EXIT:
	if (text != NULL) {
		free(text);
	}

	return result;
}
#endif

int printf_cp932(const char *format, ...)
{
	int result = -1;
	va_list ap;
	char *text = NULL;

#ifdef WIN32
	if (lang_checked == 0) {
		check_lang();
	}
#endif

	va_start(ap, format);

#ifdef WIN32
	if (cp932_conversion == 0) {
		vprintf(format, ap);
	} else {
#endif
		if ((text = make_text(format, ap)) == NULL) {
			goto ERROR_EXIT;
		}

		if (convert_and_print_text("UTF-8", "CP932", "iso-8859-1", text) < 0) {
			goto ERROR_EXIT;
		}
#ifdef WIN32
	}
#endif

	va_end(ap);

	result = 0;

ERROR_EXIT:
	if (text != NULL) {
		free(text);
	}

	return result;
}

int printf_latin1(const char *format, ...)
{
	int result = -1;
	va_list ap;
	char *text = NULL;

#ifdef WIN32
	if (lang_checked == 0) {
		check_lang();
	}
#endif

	va_start(ap, format);

	if ((text = make_text(format, ap)) == NULL) {
		goto ERROR_EXIT;
	}

	if (convert_and_print_text(latin1_to_encoding, "iso-8859-1", NULL, text) < 0) {
		goto ERROR_EXIT;
	}

	va_end(ap);

	result = 0;

ERROR_EXIT:
	if (text != NULL) {
		free(text);
	}

	return result;
}

#ifdef WIN32
static void check_lang(void)
{
	const char *lang;

	if ((lang = getenv("LANG")) == NULL) {
		latin1_to_encoding = "CP932";
		cp932_conversion = 0;
		utf8_conversion = 1;
	} else {
		char *dot = strrchr(lang, '.');
		if (dot != NULL) {
			lang = dot + 1;
		}
		if (strcasecmp(lang, "UTF-8") == 0) {
			latin1_to_encoding = "UTF-8";
			cp932_conversion = 1;
			utf8_conversion = 0;
		} else {
			latin1_to_encoding = "CP932";
			cp932_conversion = 0;
			utf8_conversion = 1;
		}
	}

	lang_checked = 1;
}
#endif

static char *make_text(const char *format, va_list ap)
{
	int result = -1;
	va_list ap2;
	int text_size = 0;
	char *text = NULL;

	va_copy(ap2, ap);

	/* 必要なバッファサイズを求める */
	if ((text_size = vsnprintf(NULL, 0, format, ap2)) <= 0) {
		va_end(ap2);
		goto ERROR_EXIT;
	}

	/* バッファ確保 */
	if ((text = (char *)malloc(text_size + 1)) == NULL) {
		goto ERROR_EXIT;
	}

	/* 文字列作成 */
	if (vsnprintf(text, text_size + 1, format, ap) != text_size) {
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (result < 0 && text != NULL) {
		free(text);
		text = NULL;
	}

	return text;
}

static int convert_and_print_text(const char *to_encoding, const char *from_encoding, const char *from_encoding2, char *text)
{
	int result = -1;
	iconv_t cd = (iconv_t)-1;
	char *src;
	size_t src_left;

	char *buffer = NULL;
	size_t buffer_size = 0;
	char *des;
	size_t des_left;

	buffer_size = 4096;
	if ((buffer = (char *)malloc(buffer_size)) == NULL) {
		printf("\n");
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	des = buffer;
	des_left = buffer_size - 1;

	/* 文字コード変換 */
	if ((cd = iconv_open(to_encoding, from_encoding)) == (iconv_t)-1) {
		printf("\n");
		printf("ERROR: %d: iconv_open(%s, %s)\n", errno, to_encoding, from_encoding);
		goto ERROR_EXIT;
	}

	src = text;
	src_left = strlen(src);
	while (0 < src_left) {
		if (iconv(cd, &src, &src_left, &des, &des_left) == (size_t)-1) {
			if (errno == E2BIG) {
				size_t len = des - buffer;

				buffer_size += 4096;
				if ((buffer = (char *)realloc(buffer, buffer_size)) == NULL) {
					printf("\n");
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}
				des = buffer + len;
				des_left = buffer_size - len - 1;

				continue;
			} else if (errno == EILSEQ && from_encoding2 != NULL) {
				/* エンコーディングを変更してやり直す */
				iconv_close(cd);

				from_encoding = from_encoding2;
				from_encoding2 = NULL;
				if ((cd = iconv_open(to_encoding, from_encoding)) == (iconv_t)-1) {
					printf("\n");
					printf("ERROR: %d: iconv_open(%s, %s)\n", errno, to_encoding, from_encoding);
					goto ERROR_EXIT;
				}

				src = text;
				src_left = strlen(src);

				des = buffer;
				des_left = buffer_size - 1;

				continue;
#ifdef WIN32
			} else if (errno == 42) {
				/*
				 * mingwのlibiconvはFreeBSDのものとは違って変換不能の文字を置換してくれないので自前で処理
				 */

				/* 変換できない文字を ? に置換 */
				*des++ = '?';
				des_left--;

				/* 置換した文字をスキップ */
				if (strcasecmp(from_encoding, "UTF-8") == 0) {
					if ((*src & 0xC0) == 0xC0) {
						src++;
						src_left--;
						while (0 < src_left && (*src & 0xC0) == 0x80) {
							src++;
							src_left--;
						}
					} else {
						src++;
						src_left--;
					}
				} else {
					src++;
					src_left--;
				}
#endif
			} else {
				printf("\n");
				printf("ERROR: %d: iconv\n", errno);
				goto ERROR_EXIT;
			}
		}
	}
	*des = '\0';
	fputs(buffer, stdout);

	result = 0;

ERROR_EXIT:
	if (cd != (iconv_t)-1) {
		iconv_close(cd);
	}
	if (buffer != NULL) {
		free(buffer);
	}

	return result;
}

/* EOF */
