/*
 * read_xrc.c
 *
 * マクロMODIFY_XRCを定義すると修正モード
 * 未定義のときは読み込みモード
 *
 * xrcファイルの読み込み処理はTLJViewを参考に作成しました
 * TLJView 1.40.7406.0 Copyright (c) 2005-2011 Deniz Oezmen (http://oezmen.eu/)
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <iconv.h>
#include <errno.h>

#include "utils.h"
#include "my_printf.h"

#ifdef MODIFY_XRC
#include "utils_utf8.h"
#endif	/* MODIFY_XRC */


#ifdef WIN32
#define printf		printf_utf8
#endif


#ifdef MODIFY_XRC
static const uint16_t WARN_TEXT_SIZE = 0x7FFF;
static const uint16_t MAX_TEXT_SIZE  = 0xFFFF;
#endif	/* MODIFY_XRC */

static const int type_ids_to_replace_name[] = {
#if 0	/* 日本語化modでは変更されるが置換しない方が良さそうに見える */
	0x03,
	0x04,
	0x07,
	0x08,
	0x09,
	0x0B,
	0x10,
	0x12,
	0x13,
	0x15,
	0x16,
	0x17,
	0x1A,
	0x1B,
	0x1D,
	0x25,
#else
	0x08,
	0x12,
	0x13,
	0x17,
	0x1B,
	0x25,
#endif
	-1
};

/*
 * type id
 * 03 ?
 * 04 ?
 * 07 ?
 * 08 object
 * 09 ?
 * 0B animation
 * 0D image
 * 0F anim script line
 * 10 sound
 * 12 移動先表示?
 * 13 ?
 * 15 script var
 * 16 ?
 * 17 ?
 * 1A ?
 * 1B ?
 * 1D subtitle
 * 20 model
 * 23 lip sync
 * 24 stock sound
 * 25 ?
 * 26 texture
 */

struct xrc_info_t {
	uint8_t				type_id;
	uint8_t				tag1;
	uint16_t			tag2;
	uint16_t			name_len;
	char				*name;
	uint32_t			data_size;
	unsigned char		*data_buffer;
	uint16_t			num_children;
	uint16_t			tag4;

	/* type_id == 0x0D && tag1 == 0x04 のときだけ使用する */
	struct xrc_unknown_0D_04_t {
		uint8_t			data_type;			/* 殆どの場合は1 */
		uint32_t		unknown1_size;		/* data_type==1のとき0x42, 2のとき0x66 */
		uint16_t		text_len;
		char			*text;
		uint32_t		unknown2_size;		/* 8 */
	} unknown_0D_04;

	/* type_id == 0x1D のときだけ使用する */
	struct xrc_subtitle_t {
		uint16_t			text_len;
		char				*text;
		uint32_t			unknown;		/* 何かの属性値? */
	} subtitle;

	struct xrc_info_t	*children;
};

struct xrc_data_t {
	unsigned char		*data;
	size_t				size;
	struct xrc_info_t	*top;
};

#ifdef MODIFY_XRC
struct text_table_t {
	int					size;
	int					used;
	int					*en_line;
	char				**en_text;
	char				**ja_text;
	int					*ja_used;
};
#endif


static void usage(void);
#ifndef MODIFY_XRC
static int read_xrc_file(const char *xrc_file, const char *text_file);
#endif	/* not MODIFY_XRC */
#ifdef MODIFY_XRC
static int read_xrc(struct xrc_info_t *xrc, unsigned char *data, size_t *offset, size_t size);
#else	/* MODIFY_XRC */
static int read_xrc(struct xrc_info_t *xrc, unsigned char *data, size_t *offset, size_t size, FILE *fp);
static void print_tabs(int num_tabs);
static int print_escape_text(const char *text, int indent, int space);
static int write_escape_text(FILE *fp, char *text);
#endif	/* MODIFY_XRC */
static int check_type_id(uint8_t type_id);
#ifdef MODIFY_XRC
static int modify_xrc_file(const char *in_file, const char *text_file, const char *out_file);
static int write_xrc_file(const char *filename, const struct xrc_info_t *xrc, struct text_table_t *text_table);
static int write_xrc(FILE *fp, const struct xrc_info_t *xrc, struct text_table_t *text_table);
static int read_text_file(const char *filename, struct text_table_t *text_table);
static int read_next_line(char **buffer, size_t *size, FILE *fp);
static void strip_trailing_crlf(char *text);
static int add_text(struct text_table_t *text_table, int en_line, const char *en_text, const char *ja_text);
static const char *find_ja_text(struct text_table_t *text_table, const char *en_text);
static void print_unused_text(const struct text_table_t *text_table);
#endif	/* MODIFY_XRC */
static int convert_text_encoding(const char *to_encoding, const char *from_encoding, const char *from_encoding2, char *text, char **new_text);
static inline int is_sjis_lead_byte(unsigned char c);
static inline int is_sjis_trail_byte(unsigned char c);
#ifdef MODIFY_XRC
static int unescape_text(char *text, int is_sjis);
#else	/* MODIFY_XRC */
static int escape_text(const char *text, char **new_text, int indent, int space, int try_sjis);
#endif	/* MODIFY_XRC */


#ifdef MODIFY_XRC
static int opt_warn_long_text = 0;
#endif	/* MODIFY_XRC */
static int opt_text_utf8 = 0;
#ifdef MODIFY_XRC
static int opt_utf8_conv = -1;
static const int UTF8_WINDOWS_TO_UNIX = 0;
static const int UTF8_UNIX_TO_WINDOWS = 1;
#else	/* MODIFY_XRC */
static int opt_sjis = 0;
#endif	/* MODIFY_XRC */


int main(int argc, char **argv)
{
	const char *xrc_file = NULL;
	const char *text_file = NULL;
#ifdef MODIFY_XRC
	const char *out_file = NULL;
#endif	/* MODIFY_XRC */

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "-?") == 0 || strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
#ifdef MODIFY_XRC
			} else if (strcmp(*argv, "-o") == 0 || strcmp(*argv, "--output") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 出力ファイル名がありません\n");
					return 1;
				}
				out_file = *argv;
#endif	/* MODIFY_XRC */
			} else if (strcmp(*argv, "-U") == 0 || strcmp(*argv, "--text-utf8") == 0) {
				opt_text_utf8 = 1;
#ifdef MODIFY_XRC
			} else if (strcmp(*argv, "--utf8-windows-to-unix") == 0) {
				opt_utf8_conv = UTF8_WINDOWS_TO_UNIX;
			} else if (strcmp(*argv, "--utf8-unix-to-windows") == 0) {
				opt_utf8_conv = UTF8_UNIX_TO_WINDOWS;
			} else if (strcmp(*argv, "-W") == 0 || strcmp(*argv, "--warn-long-text") == 0) {
				opt_warn_long_text = 1;
#else	/* MODIFY_XRC */
			} else if (strcmp(*argv, "-t") == 0 || strcmp(*argv, "--text") == 0) {
				if (*++argv == NULL) {
					printf("ERROR: 出力テキストファイル名がありません\n");
					return 1;
				}
				text_file = *argv;
			} else if (strcmp(*argv, "-S") == 0 || strcmp(*argv, "--sjis") == 0) {
				opt_sjis = 1;
#endif	/* MODIFY_XRC */
			} else {
				printf("ERROR: 不正なオプションです: %s\n", *argv);
				return 1;
			}
		} else {
			if (xrc_file == NULL) {
				xrc_file = *argv;
#ifdef MODIFY_XRC
			} else if (text_file == NULL) {
				text_file = *argv;
#endif	/* MODIFY_XRC */
			} else {
				printf("ERROR: コマンドラインパラメータが多すぎます: %s\n", *argv);
				return 1;
			}
		}
	}

	if (xrc_file == NULL) {
#if 1
		usage();
#else
		printf("ERROR: リソースファイルが指定されていません\n");
#endif
		return 1;
	}
#ifdef MODIFY_XRC
	if (text_file == NULL) {
		printf("ERROR: テキストファイルが指定されていません\n");
		return 1;
	}
	if (out_file == NULL &&
		(out_file = create_new_filename(xrc_file)) == NULL) {
		printf("ERROR: 出力ファイル名を作成できません\n");
		return 1;
	}
#endif	/* MODIFY_XRC */

	printf("リソースファイル: %s\n", xrc_file);
#ifdef MODIFY_XRC
	printf("翻訳ファイル    : %s\n", text_file);
	printf("出力ファイル    : %s\n", out_file);
	if (modify_xrc_file(xrc_file, text_file, out_file) < 0) {
		printf("ERROR: リソースファイルを変更できませんでした\n");
		return 1;
	}
#else	/* MODIFY_XRC */
	if (text_file != NULL) {
		printf("テキストファイル: %s\n", text_file);
	}
	if (read_xrc_file(xrc_file, text_file) < 0) {
		printf("ERROR: リソースファイルを読み込めませんでした\n");
		return 1;
	}
#endif	/* MODIFY_XRC */
	printf("処理完了\n\n");

	return 0;
}

static void usage(void)
{
#ifdef MODIFY_XRC
	printf("使用方法: modify_xrc [オプション] xrcファイル 翻訳ファイル\n");
	printf("機    能: xrcファイル中のテキストを変更して新しいxrcファイルを作成します\n");
	printf("オプション:\n");
	printf("    -o|--output ファイル\n");
	printf("                        指定したファイルに新しいxrcファイルを出力します\n");
	printf("                        指定のないときは元のファイル名の拡張子の前に .NEW を\n");
	printf("                        挿入した名前で作成します\n");
	printf("\n");
	printf("    -U|--text-utf8      翻訳ファイルがUTF-8でエンコードされているものとして\n");
	printf("                        扱います\n");
	printf("                        指定のないときは原文がLatin-1(iso-8859-1)、\n");
	printf("                        翻訳がShift_JISであるものとして扱います\n");
	printf("       --utf8-windows-to-unix\n");
	printf("       --utf8-unix-to-windows\n");
	printf("                        --text-utf8を指定しているとき翻訳テキストをShift_JISに\n");
	printf("                        変換する前に以下の文字を変換します\n");
	printf("                            U+FF5E(～) - U+301C(〜)\n");
	printf("                            U+2225(∥) - U+2016(‖)\n");
	printf("                            U+FF0D(－) - U+2212(−)\n");
	printf("                            U+FFE0(￠) - U+00A2(¢)\n");
	printf("                            U+FFE1(￡) - U+00A3(£)\n");
	printf("                            U+FFE2(￢) - U+00AC(¬)\n");
	printf("                            U+2015(―) - U+2014(—)\n");
	printf("                        これらの文字が変換エラーになるときはこのオプションを\n");
	printf("                        試してみてください\n");
#if 0
	printf("                        <注意> U+2014(—)からU+2015(―)への変換は行いません\n");
#endif
	printf("    -W|--warn-long-text\n");
	printf("                        テキストが%u(0x%X)バイトを越えるときに警告表示\n", WARN_TEXT_SIZE, WARN_TEXT_SIZE);
	printf("                        します\n");
	printf("                        置換可能な最大サイズは%u(0x%X)バイトです\n", MAX_TEXT_SIZE, MAX_TEXT_SIZE);
	printf("    -?|--help           使用方法を表示します\n");
#else	/* MODIFY_XRC */
	printf("使用方法: read_xrc [オプション] xrcファイル\n");
	printf("機    能: xrcファイルを読み込み字幕テキストを表示します\n");
	printf("オプション:\n");
	printf("    -S|--sjis           日本語化したxrcファイルを表示できるようにします\n");
	printf("                        <注意> 原文のままのテキストにあるアクセント記号つきの\n");
	printf("                               文字等は文字化けします\n");
	printf("    -t|--text ファイル  指定したファイルにテキストをタブ区切り形式で出力します\n");
	printf("                        modify_xrcが読み込む翻訳テキストファイルとは違います\n");
	printf("    -U|--text-utf8      --textオプションで出力するファイルの文字コードを\n");
	printf("                        UTF-8にします\n");
	printf("                        <注意> 日本語化したxrcファイルでは文字化けします\n");
	printf("                               --sjisありの場合は日本語化したテキスト、\n");
	printf("                               --sjisなしの場合は原文のままのテキストが\n");
	printf("                               文字化けします\n");
	printf("    -?|--help           使用方法を表示します\n");
	printf("\n");
#endif	/* MODIFY_XRC */
}

#ifndef MODIFY_XRC
static int read_xrc_file(const char *xrc_file, const char *text_file)
{
	int result = -1;
	FILE *fp = NULL;
	struct xrc_data_t xrc_data;
	size_t alloc_size = 0;
	size_t offset = 0;

	const char *header = "# offset\ttype_id\ttag1\ttext_type\ttext\r\n";

	memset(&xrc_data, 0, sizeof(xrc_data));

	if (read_file(xrc_file, &(xrc_data.data), &(xrc_data.size)) < 0) {
		printf("ERROR: ファイルを読み込めません: %s\n", xrc_file);
		goto ERROR_EXIT;
	}

	alloc_size = sizeof(struct xrc_info_t);
	if ((xrc_data.top = (struct xrc_info_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	memset(xrc_data.top, 0, alloc_size);

	if (text_file != NULL) {
		if ((fp = fopen(text_file, "wb")) == NULL) {
			printf("ERROR: %d: ファイルを開けません: %s\n", errno, text_file);
			goto ERROR_EXIT;
		}

		if (fprintf(fp, "%s", header) < 0) {
			printf("ERROR: %d: テキストファイルを書けません: %s\n", errno, text_file);
		}
	}

	if (read_xrc(xrc_data.top, xrc_data.data, &offset, xrc_data.size, fp) < 0) {
		printf("ERROR: データ解析できません: %s\n", xrc_file);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}
#endif	/* not MODIFY_XRC */

#ifdef MODIFY_XRC
static int read_xrc(struct xrc_info_t *xrc, unsigned char *data, size_t *offset, size_t size)
#else	/* MODIFY_XRC */
static int read_xrc(struct xrc_info_t *xrc, unsigned char *data, size_t *offset, size_t size, FILE *fp)
#endif	/* MODIFY_XRC */
{
#ifndef MODIFY_XRC
	static int level = 0;
#endif	/* not MODIFY_XRC */
	int result = -1;
	size_t start_offset = *offset;

#ifndef MODIFY_XRC
	level++;
#endif	/* not MODIFY_XRC */

	if (get_uint8 (data, offset, size, &(xrc->type_id)) < 0 ||
		get_uint8 (data, offset, size, &(xrc->tag1)) < 0 ||
		get_uint16(data, offset, size, &(xrc->tag2)) < 0 ||
		get_uint16(data, offset, size, &(xrc->name_len)) < 0 ||
		get_string(data, offset, size, &(xrc->name), xrc->name_len) < 0 ||
		get_uint32(data, offset, size, &(xrc->data_size)) < 0) {
		goto ERROR_EXIT;
	}

	xrc->data_buffer = data + *offset;
	*offset += xrc->data_size;

	if (size < *offset) {
		goto ERROR_EXIT;
	}

	if (get_uint16(data, offset, size, &(xrc->num_children)) < 0 ||
		get_uint16(data, offset, size, &(xrc->tag4)) < 0) {
		goto ERROR_EXIT;
	}

#ifndef MODIFY_XRC
	print_tabs(level);
	printf("%08zX: %02X %02X ", start_offset, xrc->type_id, xrc->tag1);
	print_escape_text(xrc->name, level, 8);
	printf("\n");

	if (fp != NULL &&
		check_type_id(xrc->type_id) != 0) {
		if (fprintf(fp, "%08zX\t%02X\t%02X\t%s\t", start_offset, xrc->type_id, xrc->tag1, "0") < 0 ||
			write_escape_text(fp, xrc->name) < 0 ||
			fprintf(fp, "\r\n") < 0) {
			printf("ERROR: %d: テキストファイルを書けません\n", errno);
			goto ERROR_EXIT;
		}
	}
#endif	/* not MODIFY_XRC */

	if (xrc->type_id == 0x0D && xrc->tag1 == 0x04) {
		/* 不明 - ポップアップテキスト??? */
		size_t offset = 0x12;

#if 0
		if (xrc->data_buffer[0x12] != 0x01) {
			printf("DEBUG: data_buffer[0x12] = 0x%02X\n", xrc->data_buffer[0x12]);
		}
#endif
		if (xrc->data_size < offset + 1) {
			printf("WARNING: データ種別を取得できません: offset=%08zX\n", start_offset);
		} else if (get_uint8(xrc->data_buffer, &offset, xrc->data_size, &(xrc->unknown_0D_04.data_type)) < 0) {
			goto ERROR_EXIT;
		}
		if (xrc->unknown_0D_04.data_type == 0x01) {
			offset = 0x42;
		} else {
			offset = 0x66;
		}
		xrc->unknown_0D_04.unknown1_size = offset;

		if (xrc->data_size < offset + 2) {
			printf("WARNING: テキストサイズを取得できません: offset=%08zX\n", start_offset);
		} else if (get_uint16(xrc->data_buffer, &offset, xrc->data_size, &(xrc->unknown_0D_04.text_len)) < 0) {
			goto ERROR_EXIT;
		}

		if (xrc->data_size < offset + xrc->unknown_0D_04.text_len) {
			printf("WARNING: テキストを取得できません: offset=%08zX\n", start_offset);
		} else if (get_string(xrc->data_buffer, &offset, xrc->data_size, &(xrc->unknown_0D_04.text), xrc->unknown_0D_04.text_len) < 0) {
			goto ERROR_EXIT;
		}

		xrc->unknown_0D_04.unknown2_size = xrc->data_size - offset;
#if 0
		if (xrc->data_size != offset + 8) {
			printf("DEBUG: データサイズが一致しません: offset=%08zX\n", start_offset);
		}
#endif

#ifndef MODIFY_XRC
		if (xrc->unknown_0D_04.text != NULL) {
			print_tabs(level);
			printf("text: ");
			print_escape_text(xrc->unknown_0D_04.text, level, 6);
			printf("\n");

			if (fp != NULL) {
				if (fprintf(fp, "%08zX\t%02X\t%02X\t%s\t", start_offset, xrc->type_id, xrc->tag1, "1") < 0 ||
					write_escape_text(fp, xrc->unknown_0D_04.text) < 0 ||
					fprintf(fp, "\r\n") < 0) {
					printf("ERROR: %d: テキストファイルを書けません\n", errno);
					goto ERROR_EXIT;
				}
			}
		}
#endif	/* not MODIFY_XRC */
	} else if (xrc->type_id == 0x1D) {
		/* 字幕テキスト */
		size_t offset = 0;

		if (get_uint16(xrc->data_buffer, &offset, xrc->data_size, &(xrc->subtitle.text_len)) < 0 ||
			get_string(xrc->data_buffer, &offset, xrc->data_size, &(xrc->subtitle.text), xrc->subtitle.text_len) < 0 ||
			get_uint32(xrc->data_buffer, &offset, xrc->data_size, &(xrc->subtitle.unknown)) < 0) {
			goto ERROR_EXIT;
		}

#ifndef MODIFY_XRC
		print_tabs(level);
		printf("subtitle: %08X ", xrc->subtitle.unknown);
		print_escape_text(xrc->subtitle.text, level, 19);
		printf("\n");

		if (fp != NULL) {
			if (fprintf(fp, "%08zX\t%02X\t%02X\t%s\t", start_offset, xrc->type_id, xrc->tag1, "1") < 0 ||
				write_escape_text(fp, xrc->subtitle.text) < 0 ||
				fprintf(fp, "\r\n") < 0) {
				printf("ERROR: %d: テキストファイルを書けません\n", errno);
				goto ERROR_EXIT;
			}
		}
#endif	/* not MODIFY_XRC */

		if (offset < xrc->data_size) {
			size_t ofs;

			printf("%08zX: unknown data:", offset);
			for (ofs = offset; ofs < xrc->data_size; ofs++) {
				printf(" %02X", xrc->data_buffer[ofs]);
			}
			printf("\n");
		}
	}

	if (xrc->num_children != 0) {
		int i;
		size_t alloc_size;

		alloc_size = sizeof(struct xrc_info_t) * (xrc->num_children);
		if ((xrc->children = (struct xrc_info_t *)malloc(alloc_size)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}
		memset(xrc->children, 0, alloc_size);

		for (i = 0; i < xrc->num_children; i++) {
#ifdef MODIFY_XRC
			if (read_xrc(&(xrc->children[i]), data, offset, size) < 0)
#else	/* MODIFY_XRC */
			if (read_xrc(&(xrc->children[i]), data, offset, size, fp) < 0)
#endif	/* MODIFY_XRC */
			{
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:
#ifndef MODIFY_XRC
	level--;
#endif	/* not MODIFY_XRC */

	return result;
}

#ifndef MODIFY_XRC
static void print_tabs(int num_tabs)
{
	while (0 < num_tabs--) {
		putchar('\t');
	}
}

static int print_escape_text(const char *text, int indent, int space)
{
	int result = -1;
	char *new_text = NULL;

	if (escape_text(text, &new_text, indent, space, opt_sjis) < 0) {
		goto ERROR_EXIT;
	}

	if (new_text != NULL) {
		text = new_text;
	}

	if (opt_sjis != 0) {
		if (printf_cp932("%s", text) < 0) {
			goto ERROR_EXIT;
		}
	} else {
		if (printf_latin1("%s", text) < 0) {
			goto ERROR_EXIT;
		}
	}

	result = 0;

ERROR_EXIT:
	if (new_text != NULL) {
		free(new_text);
	}

	return result;
}

static int write_escape_text(FILE *fp, char *text)
{
	int result = -1;
	char *new_text_utf8 = NULL;
	char *new_text = NULL;
	int try_sjis = opt_sjis;

	if (opt_text_utf8 != 0) {
		if (opt_sjis != 0) {
			/* 日本語化したテキストはShift_JIS、それ以外はiso-8859-1 */
			/* 先にShift_JISから変換し、失敗したらiso-8859-1でやり直す */
			if (convert_text_encoding("UTF-8", "CP932", "iso-8859-1", text, &new_text_utf8) < 0) {
				printf("ERROR: テキストをUTF-8に変換できません\n");
				goto ERROR_EXIT;
			}
		} else {
			if (convert_text_encoding("UTF-8", "iso-8859-1", NULL, text, &new_text_utf8) < 0) {
				printf("ERROR: テキストをUTF-8に変換できません\n");
				goto ERROR_EXIT;
			}
		}
		text = new_text_utf8;
		try_sjis = 0;
	}

	if (escape_text(text, &new_text, 0, 0, try_sjis) < 0) {
		goto ERROR_EXIT;
	}

	if (new_text != NULL) {
		text = new_text;
	}

	if (fwrite(text, strlen(text), 1, fp) != 1) {
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (new_text_utf8 != NULL) {
		free(new_text_utf8);
	}
	if (new_text != NULL) {
		free(new_text);
	}

	return result;
}
#endif	/* not MODIFY_XRC */

static int check_type_id(uint8_t type_id)
{
	int i;

	for (i = 0; 0 <= type_ids_to_replace_name[i]; i++) {
		if (type_ids_to_replace_name[i] == type_id) {
			return 1;
		}
	}

	return 0;
}

#ifdef MODIFY_XRC
static int modify_xrc_file(const char *in_file, const char *text_file, const char *out_file)
{
	int result = -1;
	struct text_table_t text_table;
	struct xrc_data_t xrc_data;
	size_t alloc_size = 0;
	size_t offset = 0;

	memset(&text_table, 0, sizeof(text_table));
	memset(&xrc_data, 0, sizeof(xrc_data));

	if (read_text_file(text_file, &text_table) < 0) {
		printf("ERROR: ファイルを読み込めません: %s\n", text_file);
		goto ERROR_EXIT;
	}

	if (text_table.used == 0) {
		printf("ERROR: テキストがありません: %s\n", text_file);
		goto ERROR_EXIT;
	}

	if (read_file(in_file, &(xrc_data.data), &(xrc_data.size)) < 0) {
		printf("ERROR: ファイルを読み込めません: %s\n", in_file);
		goto ERROR_EXIT;
	}

	alloc_size = sizeof(struct xrc_info_t);
	if ((xrc_data.top = (struct xrc_info_t *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	memset(xrc_data.top, 0, alloc_size);

	if (read_xrc(xrc_data.top, xrc_data.data, &offset, xrc_data.size) < 0) {
		printf("ERROR: データ解析できません: %s\n", in_file);
		goto ERROR_EXIT;
	}

	if (write_xrc_file(out_file, xrc_data.top, &text_table) < 0) {
		printf("ERROR: ファイルを書き込めません: %s\n", out_file);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static int write_xrc_file(const char *filename, const struct xrc_info_t *xrc, struct text_table_t *text_table)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (write_xrc(fp, xrc, text_table) < 0) {
		printf("ERROR: ファイルを書けません: %s\n", filename);
		goto ERROR_EXIT;
	}

	print_unused_text(text_table);

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int write_xrc(FILE *fp, const struct xrc_info_t *xrc, struct text_table_t *text_table)
{
	int result = -1;
	int name_modified = 0;
	int data_modified = 0;

	if (write_uint8 (fp, xrc->type_id) < 0 ||
		write_uint8 (fp, xrc->tag1) < 0 ||
		write_uint16(fp, xrc->tag2) < 0) {
		goto ERROR_EXIT;
	}

	if (check_type_id(xrc->type_id) != 0) {
		size_t name_len = 0;
		const char *name = find_ja_text(text_table, xrc->name);

		if (name != NULL) {
			name_len = strlen(name);

			if (MAX_TEXT_SIZE < name_len ||
				(WARN_TEXT_SIZE < name_len && opt_warn_long_text != 0)) {
				if (MAX_TEXT_SIZE < name_len) {
					printf("ERROR: 翻訳テキストが長すぎます: 最大%u(0x%X)バイト", MAX_TEXT_SIZE, MAX_TEXT_SIZE);
				} else {
					printf("WARNING: 翻訳テキストが%u(0x%X)バイトを越えています", WARN_TEXT_SIZE, WARN_TEXT_SIZE);
				}
				printf(": %zu(0x%zX)バイト\n", name_len, name_len);
				printf("\t原文: ");
				printf_latin1("%s\n", xrc->name);
				printf("\t翻訳: ");
				printf_cp932("%s\n", name);

				if (MAX_TEXT_SIZE < name_len) {
					goto ERROR_EXIT;
				}
			}

			if (write_uint16(fp, (uint16_t)name_len) < 0 ||
				write_buffer(fp, (const unsigned char *)name, name_len) < 0) {
				goto ERROR_EXIT;
			}

			name_modified = 1;
		}
	}

	if (name_modified == 0) {
		if (write_uint16(fp, xrc->name_len) < 0 ||
			write_buffer(fp, (const unsigned char *)xrc->name, xrc->name_len) < 0) {
			goto ERROR_EXIT;
		}
	}

	if (xrc->type_id == 0x0D && xrc->tag1 == 0x04) {
		uint32_t data_size = 0;
		size_t text_len = 0;
		const char *text = find_ja_text(text_table, xrc->unknown_0D_04.text);

		if (text != NULL) {
			text_len = strlen(text);

			if (MAX_TEXT_SIZE < text_len ||
				(WARN_TEXT_SIZE < text_len && opt_warn_long_text != 0)) {
				if (MAX_TEXT_SIZE < text_len) {
					printf("ERROR: 翻訳テキストが長すぎます: 最大%u(0x%X)バイト", MAX_TEXT_SIZE, MAX_TEXT_SIZE);
				} else {
					printf("WARNING: 翻訳テキストが%u(0x%X)バイトを越えています", WARN_TEXT_SIZE, WARN_TEXT_SIZE);
				}
				printf(": %zu(0x%zX)バイト\n", text_len, text_len);
				printf("\t原文: ");
				printf_latin1("%s\n", xrc->unknown_0D_04.text);
				printf("\t翻訳: ");
				printf_cp932("%s\n", text);

				if (MAX_TEXT_SIZE < text_len) {
					goto ERROR_EXIT;
				}
			}

			data_size = xrc->unknown_0D_04.unknown1_size + 2 + text_len + xrc->unknown_0D_04.unknown2_size;

			if (write_uint32(fp, data_size) < 0 ||
				write_buffer(fp, xrc->data_buffer, xrc->unknown_0D_04.unknown1_size) < 0 ||
				write_uint16(fp, (uint16_t)text_len) < 0 ||
				write_buffer(fp, (const unsigned char *)text, text_len) < 0 ||
				write_buffer(fp, xrc->data_buffer + xrc->unknown_0D_04.unknown1_size + 2 + xrc->unknown_0D_04.text_len, xrc->unknown_0D_04.unknown2_size) < 0) {
				goto ERROR_EXIT;
			}

			data_modified = 1;
		}
	} else if (xrc->type_id == 0x1D) {
		uint32_t data_size = 0;
		size_t text_len = 0;
		const char *text = find_ja_text(text_table, xrc->subtitle.text);

		if (text != NULL) {
			text_len = strlen(text);

			if (MAX_TEXT_SIZE < text_len ||
				(WARN_TEXT_SIZE < text_len && opt_warn_long_text != 0)) {
				if (MAX_TEXT_SIZE < text_len) {
					printf("ERROR: 翻訳テキストが長すぎます: 最大%u(0x%X)バイト", MAX_TEXT_SIZE, MAX_TEXT_SIZE);
				} else {
					printf("WARNING: 翻訳テキストが%u(0x%X)バイトを越えています", WARN_TEXT_SIZE, WARN_TEXT_SIZE);
				}
				printf(": %zu(0x%zX)バイト\n", text_len, text_len);
				printf("\t原文: ");
				printf_latin1("%s\n", xrc->subtitle.text);
				printf("\t翻訳: ");
				printf_cp932("%s\n", text);

				if (MAX_TEXT_SIZE < text_len) {
					goto ERROR_EXIT;
				}
			}

			data_size = 2 + text_len + 4;

			if (write_uint32(fp, data_size) < 0 ||
				write_uint16(fp, (uint16_t)text_len) < 0 ||
				write_buffer(fp, (const unsigned char *)text, text_len) < 0 ||
				write_uint32(fp, xrc->subtitle.unknown) < 0) {
				goto ERROR_EXIT;
			}

			data_modified = 1;
		}
	}

	if (data_modified == 0) {
		if (write_uint32(fp, xrc->data_size) < 0) {
			goto ERROR_EXIT;
		}
		if (xrc->data_size != 0) {
			if (write_buffer(fp, xrc->data_buffer, xrc->data_size) < 0) {
				goto ERROR_EXIT;
			}
		}
	}

	if (write_uint16(fp, xrc->num_children) < 0 ||
		write_uint16(fp, xrc->tag4) < 0) {
		goto ERROR_EXIT;
	}

	if (xrc->num_children != 0) {
		int i;
		for (i = 0; i < xrc->num_children; i++) {
			if (write_xrc(fp, &(xrc->children[i]), text_table) < 0) {
				goto ERROR_EXIT;
			}
		}
	}

	result = 0;

ERROR_EXIT:

	return result;
}

static int read_text_file(const char *filename, struct text_table_t *text_table)
{
	const char bom[3] = { 0xEF, 0xBB, 0xBF };

	int result = -1;
	FILE *fp = NULL;
	int line = 0;
	int read_result;

	int en_line = 0;
	char *en_buffer = NULL;
	size_t en_buffer_size = 0;

	char *en_text = NULL;
	char *new_en_text2 = NULL;

	char *ja_buffer = NULL;
	size_t ja_buffer_size = 0;

	char *ja_text = NULL;
	char *new_ja_text = NULL;
	char *new_ja_text2 = NULL;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	while (1) {
		line++;
		en_line = line;
		if ((read_result = read_next_line(&en_buffer, &en_buffer_size, fp)) < 0) {
			printf("ERROR: 原文テキストを読めません: %s(%d行目)\n", filename, line);
			goto ERROR_EXIT;
		} else if (read_result == 0) {
			break;
		}
		en_text = en_buffer;

		if (line == 1 && opt_text_utf8 != 0 && memcmp(en_text, bom, 3) == 0) {
			en_text += 3;
		}

		line++;
		if ((read_result = read_next_line(&ja_buffer, &ja_buffer_size, fp)) < 0) {
			printf("ERROR: 翻訳テキストを読めません: %s(%d行目)\n", filename, line);
			goto ERROR_EXIT;
		} else if (read_result == 0) {
			printf("ERROR: 翻訳テキストがありません: %s(%d行目)\n", filename, line);
			goto ERROR_EXIT;
		}
		ja_text = ja_buffer;

		if (*ja_text != ' ') {
			printf("ERROR: 翻訳テキストが半角空白で始まっていません: %s(%d行目)\n", filename, line);
			goto ERROR_EXIT;
		}

		strip_trailing_crlf(en_text);
		strip_trailing_crlf(ja_text);

		if (*en_text == 0) {
			printf("ERROR: 原文テキストが空文字列です: %s(%d行目)\n", filename, line - 1);
			goto ERROR_EXIT;
		}

		if (*(ja_text + 1) == 0) {
			printf("WARNING: 翻訳テキストが空文字列です: %s(%d行目)\n", filename, line);
		}

		if (opt_text_utf8 != 0) {
			if (opt_utf8_conv == UTF8_WINDOWS_TO_UNIX) {
				if (convert_utf8_windows_to_unix(ja_text, &new_ja_text) < 0) {
					printf("ERROR: 記号を変換できませんでした: %s(%d行目)\n", filename, line);
				}
				if (new_ja_text != NULL) {
					ja_text = new_ja_text;
				}
			} else if (opt_utf8_conv == UTF8_UNIX_TO_WINDOWS) {
				if (convert_utf8_unix_to_windows(ja_text, &new_ja_text) < 0) {
					printf("ERROR: 記号を変換できませんでした: %s(%d行目)\n", filename, line);
				}
				if (new_ja_text != NULL) {
					ja_text = new_ja_text;
				}
			}

			if (convert_text_encoding("iso-8859-1", "UTF-8", NULL, en_text, &new_en_text2) < 0) {
				printf("ERROR: 原文テキストをLatin-1(iso-8859-1)に変換できません: %s(%d行目)\n", filename, line - 1);
				goto ERROR_EXIT;
			}
			en_text = new_en_text2;
			if (convert_text_encoding("CP932", "UTF-8", NULL, ja_text, &new_ja_text2) < 0) {
				printf("ERROR: 翻訳テキストをCP932(≒Shift_JIS)に変換できません: %s(%d行目)\n", filename, line);
				goto ERROR_EXIT;
			}
			ja_text = new_ja_text2;
			if (new_ja_text != NULL) {
				free(new_ja_text);
				new_ja_text = NULL;
			}
		}

		if (unescape_text(en_text, 0) < 0) {
			printf("ERROR: エスケープ解除できません: %s(%d行目)\n", filename, line - 1);
			goto ERROR_EXIT;
		}
		if (unescape_text(ja_text, 1) < 0) {
			printf("ERROR: エスケープ解除できません: %s(%d行目)\n", filename, line);
			goto ERROR_EXIT;
		}

		if (add_text(text_table, en_line, en_text, ja_text + 1) < 0) {
			printf("ERROR: テキストを登録できません\n");
			goto ERROR_EXIT;
		}

		if (new_en_text2 != NULL) {
			free(new_en_text2);
			new_en_text2 = NULL;
		}

		if (new_ja_text != NULL) {
			free(new_ja_text);
			new_ja_text = NULL;
		}
		if (new_ja_text2 != NULL) {
			free(new_ja_text2);
			new_ja_text2 = NULL;
		}
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (en_buffer != NULL) {
		free(en_buffer);
	}
	if (new_en_text2 != NULL) {
		free(new_en_text2);
	}
	if (ja_buffer != NULL) {
		free(ja_buffer);
	}
	if (new_ja_text != NULL) {
		free(new_ja_text);
	}
	if (new_ja_text2 != NULL) {
		free(new_ja_text2);
	}

	return result;
}

static int read_next_line(char **buffer, size_t *size, FILE *fp)
{
	int result = -1;
	int result2 = 0;
	char *ptr = *buffer;
	const int expand_size = 10240;

	if (ptr == NULL) {
		*size = expand_size;
		if ((*buffer = ptr = (char *)malloc(*size)) == NULL) {
			printf("ERROR: memory exhausted!\n");
			goto ERROR_EXIT;
		}
	}

	if (fgets(ptr, *size, fp) != NULL) {
		size_t offset = strlen(ptr);

		while (0 < offset && ptr[offset - 1] != '\r' && ptr[offset - 1] != '\n') {
			*size += expand_size;
			if ((*buffer = ptr = (char *)realloc(ptr, *size)) == NULL) {
				printf("ERROR: memory exhausted!\n");
				goto ERROR_EXIT;
			}

			if (fgets(ptr + offset, *size - offset, fp) == NULL) {
				break;
			}

			offset += strlen(ptr + offset);
		}

		result2 = 1;
	}

	if (ferror(fp)) {
		goto ERROR_EXIT;
	}

	result = result2;

ERROR_EXIT:
	if (*buffer == NULL) {
		*size = 0;
	}

	return result;
}

static void strip_trailing_crlf(char *text)
{
	int len = strlen(text);
	while (0 < len-- && (text[len] == '\r' || text[len] == '\n')) {
		text[len] = 0;
	}
}

static int add_text(struct text_table_t *text_table, int en_line, const char *en_text, const char *ja_text)
{
	if (text_table->size <= text_table->used) {
		int new_size = text_table->size + 1024;
		size_t alloc_size = sizeof(char *) * new_size;
		size_t expand_size = sizeof(char *) * (new_size - text_table->size);

		if (text_table->size == 0) {
			text_table->en_text = (char **)malloc(alloc_size);
		} else {
			text_table->en_text = (char **)realloc(text_table->en_text, alloc_size);
		}
		if (text_table->en_text == NULL) {
			printf("ERROR: memory exhausted!\n");
			return -1;
		}
		memset(text_table->en_text + text_table->size, 0, expand_size);

		if (text_table->size == 0) {
			text_table->ja_text = (char **)malloc(alloc_size);
		} else {
			text_table->ja_text = (char **)realloc(text_table->ja_text, alloc_size);
		}
		if (text_table->ja_text == NULL) {
			printf("ERROR: memory exhausted!\n");
			return -1;
		}
		memset(text_table->ja_text + text_table->size, 0, expand_size);

		alloc_size = sizeof(int) * new_size;
		expand_size = sizeof(int) * (new_size - text_table->size);

		if (text_table->size == 0) {
			text_table->en_line = (int *)malloc(alloc_size);
		} else {
			text_table->en_line = (int *)realloc(text_table->en_line, alloc_size);
		}
		if (text_table->en_line == NULL) {
			printf("ERROR: memory exhausted!\n");
			return -1;
		}
		memset(text_table->en_line + text_table->size, 0, expand_size);

		if (text_table->size == 0) {
			text_table->ja_used = (int *)malloc(alloc_size);
		} else {
			text_table->ja_used = (int *)realloc(text_table->ja_used, alloc_size);
		}
		if (text_table->ja_used == NULL) {
			printf("ERROR: memory exhausted!\n");
			return -1;
		}
		memset(text_table->ja_used + text_table->size, 0, expand_size);

		text_table->size = new_size;
	}

	text_table->en_line[text_table->used] = en_line;

	if ((text_table->en_text[text_table->used] = strdup(en_text)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	if ((text_table->ja_text[text_table->used] = strdup(ja_text)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	text_table->used++;

	return 0;
}

static const char *find_ja_text(struct text_table_t *text_table, const char *en_text)
{
	int i;

	if (en_text == NULL || *en_text == '\0') {
		return NULL;
	}

	for (i = 0; i < text_table->used; i++) {
		if (text_table->ja_used[i] == 0 &&
			strcmp(text_table->en_text[i], en_text) == 0) {
			text_table->ja_used[i] = 1;
			return text_table->ja_text[i];
		}
	}

	return NULL;
}

static void print_unused_text(const struct text_table_t *text_table)
{
	int count = 0;
	int i;

	for (i = 0; i < text_table->used; i++) {
		if (text_table->ja_used[i] == 0) {
			count++;
		}
	}
	printf("置換件数: %d / %d\n", text_table->used - count, text_table->used);

	count = 0;
	for (i = 0; i < text_table->used; i++) {
		if (text_table->ja_used[i] == 0) {
			if (count == 0) {
				printf("未使用テキスト:\n");
			}
			printf_latin1("\t%d\t%s\n", text_table->en_line[i], text_table->en_text[i]);
			count++;
		}
	}
}
#endif	/* MODIFY_XRC */

static int convert_text_encoding(const char *to_encoding, const char *from_encoding, const char *from_encoding2, char *text, char **new_text)
{
	int result = -1;
	iconv_t cd = (iconv_t)-1;
	char *src;
	size_t src_left;
	char *des;
	size_t des_left;
	size_t alloc_size = 4096;

	*new_text = NULL;

	if ((*new_text = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	/* 文字コード変換 */
	if ((cd = iconv_open(to_encoding, from_encoding)) == (iconv_t)-1) {
		printf("ERROR: %d: iconv_open(%s, %s)\n", errno, to_encoding, from_encoding);
		goto ERROR_EXIT;
	}

	src = text;
	src_left = strlen(src);
	des = *new_text;
	des_left = alloc_size - 1;

	while (0 < src_left) {
		if (iconv(cd, &src, &src_left, &des, &des_left) == (size_t)-1) {
			if (errno == E2BIG) {
				size_t len = des - *new_text;

				alloc_size += 4096;
				if ((*new_text = (char *)realloc(*new_text, alloc_size)) == NULL) {
					printf("ERROR: memory exhausted!\n");
					goto ERROR_EXIT;
				}
				des = *new_text + len;
				des_left = alloc_size - len - 1;

				continue;
			} else if (errno == EILSEQ && from_encoding2 != NULL) {
				/* エンコーディングを変更してやり直す */
				iconv_close(cd);

				from_encoding = from_encoding2;
				from_encoding2 = NULL;
				if ((cd = iconv_open(to_encoding, from_encoding)) == (iconv_t)-1) {
					printf("ERROR: %d: iconv_open(%s, %s)\n", errno, to_encoding, from_encoding);
					goto ERROR_EXIT;
				}

				src = text;
				src_left = strlen(src);

				des = *new_text;
				des_left = alloc_size - 1;

				continue;
#ifdef WIN32
			} else if (errno == 42) {
				/* 変換できない文字を ? に置換 */
				*des++ = '?';
				des_left--;

				/* 置換した文字をスキップ */
				if (strcasecmp(from_encoding, "UTF-8") == 0) {
					if ((*src & 0xC0) == 0xC0) {
						src++;
						src_left--;
						while (0 < src_left && (*src & 0xC0) == 0x80) {
							src++;
							src_left--;
						}
					} else {
						src++;
						src_left--;
					}
				} else {
					src++;
					src_left--;
				}
#endif
			} else {
				printf("ERROR: %d: iconv\n", errno);
				goto ERROR_EXIT;
			}
		}
	}
	*des = '\0';

	/* 文字列長に合わせてバッファサイズを変更 */
	if ((*new_text = (char *)realloc(*new_text, strlen(*new_text) + 1)) == NULL) {
		/* ここでエラーになることは無いと思う */
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (cd != (iconv_t)-1) {
		iconv_close(cd);
	}
	if (result < 0) {
		if (*new_text != NULL) {
			free(*new_text);
			*new_text = NULL;
		}
	}

	return result;
}

static inline int is_sjis_lead_byte(unsigned char c)
{
	return ((0x81 <= c && c <= 0x9F) || (0xE0 <= c && c <= 0xFC));
}

static inline int is_sjis_trail_byte(unsigned char c)
{
	return ((0x40 <= c && c <= 0x7E) || (0x80 <= c && c <= 0xFC));
}

#ifdef MODIFY_XRC
static int unescape_text(char *text, int is_sjis)
{
	char *src = text;
	char *des = text;

	while (*src != '\0') {
		if (*src == '\\') {
			src++;
			if (*src == '\\') {
				*des++ = *src++;
			} else if (*src == 't') {
				src++;
				*des++ = '\t';
			} else if (*src == 'r') {
				src++;
				*des++ = '\r';
			} else if (*src == 'n') {
				src++;
				*des++ = '\n';
			} else {
				/* 何もしない */
			}
		} else if (is_sjis && is_sjis_lead_byte(*(unsigned char *)src)) {
			*des++ = *src++;
			if (! is_sjis_trail_byte(*(unsigned char *)src)) {
				printf("ERROR: 不正なShift_JIS文字列です\n");
				return -1;
			}
			*des++ = *src++;
		} else {
			*des++ = *src++;
		}
	}
	*des = 0;

	return 0;
}
#else	/* MODIFY_XRC */
static int escape_text(const char *text, char **new_text, int indent, int space, int try_sjis)
{
	int found = 0;
	const char *src = text;
	char *des = NULL;
	size_t alloc_size = 0;

	*new_text = NULL;

	while (*src != '\0') {
		if (*src == '\\' || *src == '\t' || *src == '\r') {
			found++;
			src++;
		} else if (*src == '\n') {
			found++;
			src++;
			if ((indent != 0 || space != 0) && *src != '\0') {
				/* 表示用の改行とインデントの分を追加 */
				found += 1 + indent + space;
			}
		} else if (try_sjis && is_sjis_lead_byte(*(unsigned char *)src)) {
			src++;
			if (is_sjis_trail_byte(*(unsigned char *)src)) {
				src++;
			} else {
				/*
				 * Shift_JISのコード範囲外の文字が見つかった場合
				 * Shift_JIS指定を解除して最初から調べなおす
				 */
				try_sjis = 0;
				found = 0;
				src = text;
				continue;
			}
		} else {
			src++;
		}
	}

	if (found == 0) {
		return 0;
	}

	alloc_size = strlen(text) + found + 1;
	if ((*new_text = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}

	src = text;
	des = *new_text;

	while (*src != '\0') {
		if (*src == '\\') {
			src++;
			*des++ = '\\';
			*des++ = '\\';
		} else if (*src == '\t') {
			src++;
			*des++ = '\\';
			*des++ = 't';
		} else if (*src == '\r') {
			src++;
			*des++ = '\\';
			*des++ = 'r';
		} else if (*src == '\n') {
			src++;
			*des++ = '\\';
			*des++ = 'n';
			if ((indent != 0 || space != 0) && *src != '\0') {
				int len;

				*des++ = '\n';

				len = indent;
				while (0 < len--) {
					*des++ = '\t';
				}

				len = space;
				while (0 < len--) {
					*des++ = ' ';
				}
			}
		} else if (try_sjis && is_sjis_lead_byte(*(unsigned char *)src)) {
			*des++ = *src++;
			*des++ = *src++;
		} else {
			*des++ = *src++;
		}
	}
	*des = '\0';

	return 1;
}
#endif	/* MODIFY_XRC */

/* EOF */
