#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "utils.h"

#ifdef WIN32
#include "my_printf.h"
#define printf		printf_utf8
#endif


int get_uint8(const unsigned char *data, size_t *offset, size_t size, uint8_t *value)
{
	if (size < *offset + 1) {
		return -1;
	}

	data += *offset;
	*value = (uint8_t)*data;
	*offset += 1;

	return 0;
}

int get_uint16(const unsigned char *data, size_t *offset, size_t size, uint16_t *value)
{
	if (size < *offset + 2) {
		return -1;
	}

	data += *offset;
	*value = ((uint16_t)*data) | (((uint16_t)*(data + 1)) << 8);
	*offset += 2;

	return 0;
}

int get_uint32(const unsigned char *data, size_t *offset, size_t size, uint32_t *value)
{
	if (size < *offset + 4) {
		return -1;
	}

	data += *offset;
	*value = ((uint32_t)*data) | (((uint32_t)*(data + 1)) << 8) | (((uint32_t)*(data + 2)) << 16) | (((uint32_t)*(data + 3)) << 24);
	*offset += 4;

	return 0;
}

int get_string(const unsigned char *data, size_t *offset, size_t size, char **value, size_t len)
{
	if (size < *offset + len) {
		return -1;
	}

	if (len == 0) {
		return 0;
	}

	data += *offset;

	if ((*value = (char *)malloc(len + 1)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return -1;
	}
	memcpy(*value, data, len);
	(*value)[len] = 0;

	*offset += len;

	return len;
}

int write_uint8(FILE *fp, uint8_t value)
{
	unsigned char buffer[1];

	buffer[0] = (unsigned char)value;

	return write_buffer(fp, buffer, 1);
}

int write_uint16(FILE *fp, uint16_t value)
{
	unsigned char buffer[2];

	buffer[0] = (unsigned char)(value & 0xFF);
	value >>= 8;
	buffer[1] = (unsigned char)(value & 0xFF);

	return write_buffer(fp, buffer, 2);
}

int write_uint32(FILE *fp, uint32_t value)
{
	unsigned char buffer[4];

	buffer[0] = (unsigned char)(value & 0xFF);
	value >>= 8;
	buffer[1] = (unsigned char)(value & 0xFF);
	value >>= 8;
	buffer[2] = (unsigned char)(value & 0xFF);
	value >>= 8;
	buffer[3] = (unsigned char)(value & 0xFF);

	return write_buffer(fp, buffer, 4);
}

int write_buffer(FILE *fp, const unsigned char *buffer, size_t size)
{
	if (fwrite(buffer, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを書けません\n", errno);
		return -1;
	}

	return 0;
}

int make_directory(const char *path, int path_is_file)
{
	int result = -1;
	char *work_buffer = NULL;
	char *dir = NULL;
	char *name = NULL;
	int is_fullpath = 0;
	struct stat st;
#ifdef WIN32
	const char *STR_PATH_SEP = "\\";
	const char CHAR_PATH_SEP = '\\';
#else
	const char *STR_PATH_SEP = "/";
	const char CHAR_PATH_SEP = '/';
#endif


	if (path_is_file == 0) {
		if (stat(path, &st) == 0 && S_ISDIR(st.st_mode)) {
			goto SUCCESS_EXIT;
		}
	}

	if ((work_buffer = strdup(path)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

#ifdef WIN32
	convert_slash_to_backslash(work_buffer);
#else
	convert_backslash_to_slash(work_buffer);
#endif

	if (path_is_file) {
		char *ptr;

		if ((ptr = strrchr(work_buffer, CHAR_PATH_SEP)) == NULL || ptr == work_buffer) {
			goto SUCCESS_EXIT;
		}

		*ptr = 0;
#ifdef WIN32
		if (strlen(work_buffer) == 2 && work_buffer[1] == ':') {
			goto SUCCESS_EXIT;
		}
#endif

		if (stat(work_buffer, &st) == 0 && S_ISDIR(st.st_mode)) {
			goto SUCCESS_EXIT;
		}
	}

	if ((dir = (char *)malloc(strlen(path) + 1)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}
	dir[0] = 0;

	if (work_buffer[0] == CHAR_PATH_SEP) {
		is_fullpath = 1;
	}

	name = strtok(work_buffer, STR_PATH_SEP);
	if (name == NULL) {
		printf("ERROR: illegal path: %s\n", path);
		goto ERROR_EXIT;
	}

	while (name != NULL) {
		if (dir[0] == 0 && is_fullpath == 0) {
			strcpy(dir, name);
		} else {
			strcat(strcat(dir, STR_PATH_SEP), name);
		}

#ifdef WIN32
		if (strlen(dir) == 2 && dir[1] == ':') {
			continue;
		}
#endif
		if (stat(dir, &st) != 0) {
			int res;
#ifdef WIN32
			res = mkdir(dir);
#else
			res = mkdir(dir, 0700);
#endif
			if (res < 0) {
				printf("ERROR: %d: can't make directory: %s\n", errno, dir);
				goto ERROR_EXIT;
			}
		}

		name = strtok(NULL, STR_PATH_SEP);
	}

SUCCESS_EXIT:
	result = 0;

ERROR_EXIT:
	if (work_buffer != NULL) {
		free(work_buffer);
	}
	return result;
}

void convert_slash_to_backslash(char *str)
{
	if (str == NULL) {
		return;
	}

	while (*str != 0) {
		if (*str == '/') {
			*str = '\\';
		}
		str++;
	}
}

void convert_backslash_to_slash(char *str)
{
	if (str == NULL) {
		return;
	}

	while (*str != 0) {
		if (*str == '\\') {
			*str = '/';
		}
		str++;
	}
}

char *create_new_filename(const char *filename)
{
	size_t alloc_size = strlen(filename) + 4 + 1;
	char *new_name = NULL;
	const char *ptr = NULL;

	if ((new_name = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		return NULL;
	}

	if ((ptr = strrchr(filename, '.')) != NULL) {
		if (strchr(ptr + 1, '/') != NULL) {
			ptr = NULL;
		} else if (strchr(ptr + 1, '\\') != NULL) {
			ptr = NULL;
		}
	}

	if (ptr == NULL) {
		strcat(strcpy(new_name, filename), ".NEW");
	} else {
		int base_len = ptr - filename;
		strncpy(new_name, filename, base_len);
		strcat(strcpy(new_name + base_len, ".NEW"), ptr);
	}

	return new_name;
}

int read_file(const char *filename, unsigned char **data, size_t *size)
{
	int result = -1;
	struct stat st;
	FILE *fp = NULL;

	*data = NULL;
	*size = 0;

	if (stat(filename, &st) < 0) {
		printf("ERROR: %d: ファイル情報を取得できません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	*size = st.st_size;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if ((*data = (unsigned char *)malloc(*size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	if (fread((*data), *size, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを読めません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (result < 0) {
		if (*data != NULL) {
			free(*data);
		}
		*data = NULL;
		*size = 0;
	}

	return result;
}

int write_file(const char *filename, const unsigned char *data, size_t size)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "wb")) == NULL) {
		printf("ERROR: %d: ファイルを開けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	if (fwrite(data, size, 1, fp) != 1) {
		printf("ERROR: %d: ファイルを書けません: %s\n", errno, filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

/* EOF */
