#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "utils_utf8.h"

#ifdef WIN32
#include "my_printf.h"
#define printf		printf_utf8
#endif


static const struct cp932_ucs_table_t {
	uint16_t	cp932;
	uint32_t	ucs_windows;
	const char	*utf8_windows;
	uint32_t	ucs_unix;
	const char	*utf8_unix;
	int			diff_utf8_size;
	int			windows_to_unix_only;
} cp932_ucs_table[] = {
	{ 0x8160, 0xFF5E, "\xEF\xBD\x9E", 0x301C, "\xE3\x80\x9C", 0, 0 },		/* ～ FULLWIDTH TILDE        / 〜 WAVE DASH            */
	{ 0x6161, 0x2225, "\xE2\x88\xA5", 0x2016, "\xE2\x80\x96", 0, 0 },		/* ∥ PARALLEL TO            / ‖ DOUBLE VERTICAL LINE */
	{ 0x817C, 0xFF0D, "\xEF\xBC\x8D", 0x2212, "\xE2\x88\x92", 0, 0 },		/* － FULLWIDTH HYPHEN-MINUS / − MINUS SIGN           */
	{ 0x8191, 0xFFE0, "\xEF\xBF\xA0", 0x00A2, "\xC2\xA2"    , 1, 0 },		/* ￠ FULLWIDTH CENT SIGN    / ¢ CENT SIGN            */
	{ 0x8192, 0xFFE1, "\xEF\xBF\xA1", 0x00A3, "\xC2\xA3"    , 1, 0 },		/* ￡ FULLWIDTH POUND SIGN   / £ POUND SIGN           */
	{ 0x81CA, 0xFFE2, "\xEF\xBF\xA2", 0x00AC, "\xC2\xAC"    , 1, 0 },		/* ￢ FULLWIDTH NOT SIGN     / ¬ NOT SIGN             */

	/* Windows Vista以降ではU+2014と対応 */
	{ 0x815C, 0x2015, "\xE2\x80\x95", 0x2014, "\xE2\x80\x94", 0, 1 },		/* ― HORIZONTAL BAR         / — EM DASH              */

	{ 0, 0, NULL, 0, NULL, 0, 0 }
};


static int convert_utf8(const char *text, char **new_text, int windows_to_unix);


int convert_utf8_windows_to_unix(const char *text, char **new_text)
{
	return convert_utf8(text, new_text, 1);
}

int convert_utf8_unix_to_windows(const char *text, char **new_text)
{
	return convert_utf8(text, new_text, 0);
}

static int convert_utf8(const char *text, char **new_text, int windows_to_unix)
{
	int result = -1;

	int found = 0;
	int diff = 0;

	size_t alloc_size = 0;
	const char *src = NULL;
	char *des = NULL;

	*new_text = NULL;

	src = text;
	while (*src != '\0') {
		uint32_t code;
		int len;
		int i;

		if ((len = get_next_utf8_char(&code, src)) < 0) {
			goto ERROR_EXIT;
		} else if (len == 0) {
			/* ここには来ない */
			break;
		}

		for (i = 0; cp932_ucs_table[i].cp932 != 0; i++) {
			if (windows_to_unix) {
				if (cp932_ucs_table[i].ucs_windows == code) {
					found++;
					diff -= cp932_ucs_table[i].diff_utf8_size;
					break;
				}
			} else {
				if (cp932_ucs_table[i].ucs_unix == code) {
#if 0
					if (cp932_ucs_table[i].windows_to_unix_only == 0) {
#endif
						found++;
						diff += cp932_ucs_table[i].diff_utf8_size;
						break;
#if 0
					}
#endif
				}
			}
		}

		src += len;
	}

	if (found == 0) {
		return 0;
	}

	alloc_size = strlen(text) + 1;
	if (0 < diff) {
		alloc_size += diff;
	}

	if ((*new_text = (char *)malloc(alloc_size)) == NULL) {
		printf("ERROR: memory exhausted!\n");
		goto ERROR_EXIT;
	}

	src = text;
	des = *new_text;
	while (0 < found) {
		const char *src2 = NULL;
		uint32_t code;
		int len;
		int i;

		if ((len = get_next_utf8_char(&code, src)) < 0) {
			goto ERROR_EXIT;
		} else if (len == 0) {
			/* ここには来ない */
			break;
		}

		for (i = 0; cp932_ucs_table[i].cp932 != 0; i++) {
			if (windows_to_unix) {
				if (cp932_ucs_table[i].ucs_windows == code) {
					src2 = cp932_ucs_table[i].utf8_unix;
					break;
				}
			} else {
				if (cp932_ucs_table[i].ucs_unix == code) {
					if (cp932_ucs_table[i].windows_to_unix_only == 0) {
						src2 = cp932_ucs_table[i].utf8_windows;
						break;
					}
				}
			}
		}

		if (src2 != NULL) {
			found--;
			src += len;
			while (*src2 != '\0') {
				*des++ = *src2++;
			}
		} else {
			while (0 < len--) {
				*des++ = *src++;
			}
		}
	}

	strcpy(des, src);

	result = 1;

ERROR_EXIT:
	if (result < 0) {
		if (*new_text != NULL) {
			free(*new_text);
			*new_text = NULL;
		}
	}

	return result;
}

int get_next_utf8_char(uint32_t *code, const char *src)
{
	int len;
	int l;
	const char *save = src;
	int is_illegal = 0;

	*code = 0;

	if (src == NULL) {
		return 0;
	}

	/* 1文字分をunsigned longにパック */
	*code = (uint32_t)*src;

	if (*src == '\0') {
		return 0;
	} else if ((*src & 0x80) == 0) {
		len = 1;
	} else if ((*src & 0x40) == 0) {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです: %02X\n", (int)*(unsigned char *)src);
		return -1;
	} else if ((*src & 0x20) == 0) {
		len = 2;
		*code &= 0x1F;
	} else if ((*src & 0x10) == 0) {
		len = 3;
		*code &= 0x0F;
	} else if ((*src & 0x08) == 0) {
		len = 4;
		*code &= 0x07;
	} else if ((*src & 0x04) == 0) {
		len = 5;
		*code &= 0x03;
	} else if ((*src & 0x02) == 0) {
		len = 6;
		*code &= 0x01;
	} else {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです: %02X\n", (int)*(unsigned char *)src);
		return -1;
	}
	src++;

	l = len - 1;
	while (0 < l--) {
		*code <<= 6;
		if ((*src & 0xC0) != 0x80) {
			/* illegal UTF-8 sequence */
			is_illegal = 1;
			break;
		}
		*code |= ((unsigned char)*src++) & 0x3F;
	}

	if (is_illegal ||
		(0xD800 <= *code && *code <= 0xDFFF) ||
		(len == 2 && *code <= 0x0000007F) ||
		(len == 3 && *code <= 0x000007FF) ||
		(len == 4 && *code <= 0x0000FFFF) ||
		(len == 5 && *code <= 0x001FFFFF) ||
		(len == 6 && *code <= 0x03FFFFFF)) {
		/* illegal UTF-8 sequence */
		printf("ERROR: 不正な文字コードです:");
		while (0 < len--) {
			printf(" %02X", (int)*(unsigned char *)save);
			if (*save == '\0') {
				break;
			}
			save++;
		}
		printf("\n");
		return -1;
	}

	return len;
}

char *utf32_to_utf8(char *des, int *size, uint32_t code)
{
	int len;

	if (*size <= 0) {
		printf("BUG: code_to_string: buffer size=%d\n", *size);
		return NULL;
	}

	if (0x0000D800 <= code && code <= 0x0000DFFF) {
		printf("ERROR: 不正な文字コードです: %08X\n", code);
		return NULL;
	} else if (code < 0x0000007F) {
		len = 1;
		*des++ = (char)(code & 0x7F);
	} else if (code < 0x000007FF) {
		len = 2;
		*des++ = (char)(0xC0 | (code >> 6));
	} else if (code < 0x0000FFFF) {
		len = 3;
		*des++ = (char)(0xE0 | (code >> 12));
	} else if (code < 0x001FFFFF) {
		len = 4;
		*des++ = (char)(0xF0 | (code >> 18));
	} else if (code < 0x03FFFFFF) {
		len = 5;
		*des++ = (char)(0xF8 | (code >> 24));
	} else if (code < 0x7FFFFFFF) {
		len = 6;
		*des++ = (char)(0xFC | (code >> 30));
	} else {
		printf("ERROR: 不正な文字コードです: %08X\n", code);
		return NULL;
	}

	if (*size <= len) {
		printf("BUG: ucs_to_utf8: buffer size=%d, string length=%d\n", *size, len);
		return NULL;
	}
	*size -= len;

	len--;
	while (0 < len--) {
		*des++ = (char)(0x80 | ((code >> (6 * len)) & 0x3F));
	}
	*des = 0;

	return des;
}

/* EOF */
